#include <boost/algorithm/string/case_conv.hpp>
#include <exception>
#include <stdexcept>
#include <cstdint>
#include <string>
#include <cmath>
#include <filesystem>
#include "parsers.hpp"

namespace fs = std::filesystem;

namespace parsers {

uint64_t
parse_size(const std::string& str) {

    constexpr const uint64_t B_FACTOR = 1;
    constexpr const uint64_t KB_FACTOR = 1e3;
    constexpr const uint64_t KiB_FACTOR = (1 << 10);
    constexpr const uint64_t MB_FACTOR = 1e6;
    constexpr const uint64_t MiB_FACTOR = (1 << 20);
    constexpr const uint64_t GB_FACTOR = 1e9;
    constexpr const uint64_t GiB_FACTOR = (1 << 30);

    const std::pair<const std::string, const uint64_t> conversions[] = {
            {"GiB", GiB_FACTOR}, {"GB", GB_FACTOR}, {"G", GB_FACTOR},
            {"MiB", MiB_FACTOR}, {"MB", MB_FACTOR}, {"M", MB_FACTOR},
            {"KiB", KiB_FACTOR}, {"KB", KB_FACTOR}, {"K", KB_FACTOR},
            {"B", B_FACTOR},
    };

    std::string scopy(str);

    /* remove whitespaces from the string */
    scopy.erase(std::remove_if(scopy.begin(), scopy.end(),
                               [](char ch) {
                                 return std::isspace<char>(
                                         ch, std::locale::classic());
                               }),
                scopy.end());

    /* determine the units */
    std::size_t pos = std::string::npos;
    uint64_t factor = B_FACTOR;

    for(const auto& c : conversions) {
        const std::string& suffix = c.first;

        if((pos = scopy.find(suffix)) != std::string::npos) {
            /* check that the candidate is using the suffix EXACTLY
             * to prevent accepting something like "GBfoo" as a valid Gigabyte
             * unit */
            if(suffix != scopy.substr(pos)) {
                pos = std::string::npos;
                continue;
            }

            factor = c.second;
            break;
        }
    }

    /* this works as expected because if pos == std::string::npos, the standard
     * states that all characters until the end of the string should be
     * included.*/
    std::string number_str = scopy.substr(0, pos);

    /* check if it's a valid number */
    if(number_str.empty() ||
       !std::all_of(number_str.begin(), number_str.end(), ::isdigit)) {
        throw std::invalid_argument("Not a number");
    }

    double value = std::stod(number_str);

    return std::round(value * factor);
}

bool
parse_bool(const std::string& name, const std::string& value) {

    if(value == "1" || boost::algorithm::to_lower_copy(value) == "true") {
        return true;
    }

    if(value == "0" || boost::algorithm::to_lower_copy(value) == "false") {
        return false;
    }

    throw std::invalid_argument("Value provided for option '" + name +
                                "' is not boolean");
}

uint32_t
parse_number(const std::string& name, const std::string& value) {

    int32_t optval = 0;

    try {
        optval = std::stoi(value);
    } catch(...) {
        throw std::invalid_argument("Value provided for option '" + name +
                                    "' is not a number");
    }

    if(optval <= 0) {
        throw std::invalid_argument("Value provided for option '" + name +
                                    "' must be greater than zero");
    }

    return static_cast<uint32_t>(optval);
}

fs::path
parse_path(const std::string& name, const std::string& value) {

    (void) name;

    return {value};
}

fs::path
parse_existing_path(const std::string& name, const std::string& value) {

    if(!fs::exists(value)) {
        throw std::invalid_argument("Path '" + value + "' in option '" + name +
                                    "' does not exist");
    }

    return {value};
}

uint64_t
parse_capacity(const std::string& name, const std::string& value) {

    try {
        return parse_size(value);
    } catch(const std::exception& e) {
        throw std::invalid_argument("Value provided in option '" + name +
                                    "' is invalid");
    }
}

} // namespace parsers
